import autofit as af

# In this chapter, we're not going to individually import every module we're used to importing. Instead, we can just
# import autolens (as 'al'), which is what we'll do to import all the modules we're used to from chapter 1. However,
# we'll indivdually import the modules new to this chapter (and we'll import them as 'al' in chapter 3).

import autolens as al
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging

# In this example, we're going to find a lens model that provides a good fit to an image, without
# assuming any knowledge of what the 'correct' lens model is.

# So, whats a 'lens model'? It is the combination of light profiles and mass profiles we use to represent a lens galaxy,
# source galaxy and therefore the strong lens ray-tracing configuration (i.e. a tracer).

# To begin, we have to choose the parametrization of our lens model. We don't need to specify the values of its light
# and mass profiles (e.g. the centre, einstein_radius, etc.) - only the profiles themselves. In this example, we'll use
# the following lens model:

# 1) A spherical Isothermal Sphere (SIS) for the lens galaxy's mass.
# 2) A spherical exponential light profile for the source galaxy's light.

# I'll let you into a secret - this is the same lens model used to Simulate the CCD data we're going to fit and we're
# going to infer the actual parameters I used.

# So, how do we infer the light and mass profile parameters that give a good fit to our data?

# Well, we could randomly guess a lens model, corresponding to some random set of parameters. We could use this lens
# model to create a tracer and fit the image-data, and quantify how good the fit was using its likelihood
# (recall chapter_1/tutorial_8). If we kept guessing lens models, eventually we'd find one that provides a good fit
# (i.e. high likelihood) to the data!

# It may sound surprising, but this is actually the basis of how lens modeling works. However, we can do a lot better
# than random guessing. Instead, we track the likelihood of our previous guesses and guess more models using combinations
# of parameters that gave higher likelihood solutions previously. The idea is that if a set of parameters provided a
# good fit to the data, another set with similar values probably will too.

# This is called a 'non-linear search' and its a fairly common problem faced by scientists. Over the next few tutorials,
# we're going to really get our heads around the concept of a non-linear search - intuition which will prove crucial
# to being a successful lens modeler.

# We're going to use a non-linear search algorithm called 'MultiNest'. I highly recommend it, and find its great for lens
# modeling. However, for now, lets not worry about the details of how MultiNest actually works. Instead, just
# picture that a non-linear search in PyAutoLens operates as follows:

# 1) Randomly guess a lens model and use its light-profiles and mass-profiles to set up a lens galaxy, source galaxy
#    and a tracer.

# 2) Pass this tracer through the fitting module, generating a model image and comparing this model image to the
#    observed strong lens imaging data. This computes a likelihood.

# 3) Repeat this many times, using the likelihoods of previous fits (typically those with a high likelihood) to
#    guide us to the lens models with the highest likelihood.


# You're going to see a line like the one below (with 'conf.instance') in every tutorial this chapter. I recommend that
# for now you just ignore it. A non-linear search can take a long time to run (minutes, hours, or days), and this
# isn't ideal if you want to go through the tutorials without having to constant stop for long periods!

# This line overrides the configuration of the non-linear search such that it computes the solution really fast. To
# do this, I've 'cheated' - I've computed the solution myself and then input it into the config. Nevertheless, it means
# things will run fast for you, meaning you won't suffer long delays doing the tutorials.
#
# This will all become clear at the end of the chapter, so for now just bare in mind that we are taking a short-cut
# to get our non-linear search to run fast!

# You need to change the path below to the chapter 1 directory.
chapter_path = "/path/to/user/autolens_workspace/howtolens/chapter_2_lens_modeling/"
chapter_path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/howtolens/chapter_2_lens_modeling/"

af.conf.instance = af.conf.Config(
    config_path=chapter_path + "/configs/t1_non_linear_search",
    output_path=chapter_path + "/output",
)

# This function simulates the image we'll fit in this tutorial.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.1, pixel_scale=0.1)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(130, 130), pixel_scale=0.1, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.SphericalExponential(
            centre=(0.0, 0.0), intensity=0.2, effective_radius=0.2
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.1,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Lets call the simulate function above, providing us with CCD data to fit.
ccd_data = simulate()

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# To compute a lens model, we use a GalaxyModel, which behaves analogously to the Galaxy objects we're now used to.
# However, whereas for a Galaxy we manually specified the value of every parameter of its light-profiles and mass-profiles,
# for a GalaxyModel these are inferred by the non-linear search.

# Lets model the lens galaxy with an SIS mass profile (which is what it was simulated with).
lens_galaxy_model = gm.GalaxyModel(
    redshift=0.5, mass=al.mass_profiles.SphericalIsothermal
)

# Lets model the source galaxy with a spherical exponential light profile (again, what it was simulated with).
source_galaxy_model = gm.GalaxyModel(
    redshift=1.0, light=al.light_profiles.SphericalExponential
)

# A phase takes our galaxy models and fits their parameters via a non-linear search (in this case, MultiNest).

# (Just like we could give profiles descriptive names, like 'light', 'bulge' and 'disk', we can do the exact same
# thing with the phase's galaxies. This is good practise - as once we start using complex lens models, you could
# potentially have a lot of galaxies - and this is the best way to keep track of them!).

# (ignore the 'dict' - its necessary syntax but not something you need to concern yourself with).

phase = phase_imaging.PhaseImaging(
    phase_name="phase_t1_non_linear_search",
    galaxies=dict(lens_galaxy=lens_galaxy_model, source_galaxy=source_galaxy_model),
    optimizer_class=af.MultiNest,
)

# To run the phase, we pass it the data we're going to fit a lens model to and the non-linear search begins! As the phase
# runs, a logger will show you the parameters of the best-fit model.
print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t1_non_linear_search"
    "folder for live output of the results, images and lens model."
    "This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)

results = phase.run(data=ccd_data)

print("MultiNest has finished run - you may now continue the notebook.")

# Now this is running you should checkout the 'autolens_workspace/howtolens/chapter_2_lens_modeling/output' folder.

# This is where the results of the phase are written to your hard-disk (in the 't1_non_linear_search' folder).
# When its completed, images and output will also appear in this folder, meaning that you don't need to keep running
# python code to see the results.

# The best-fit solution (i.e. the highest likelihood) is stored in the 'results', which we can plot as per usual.
al.lens_fit_plotters.plot_fit_subplot(fit=results.most_likely_fit)

# The fit looks good and we've therefore found a model close to the one I used to simulat the image with (you can
# confirm this yourself if you want, by comparing the inferred parameters to those found in the simulate function above).

# And with that, we're done - you've successfully modeled your first strong lens with PyAutoLens! Before moving
# onto the next tutorial, I want you to think about the following:

# 1) a non-linear search is often said to search a 'non-linear parameter-space' - why is the term parameter-space used?

# 2) Why is this parameter space 'non-linear'?

# 3) Initially, the non-linear search randomly guesses the values of the parameters. However, it shouldn't 'know' what
#    reasonable values for a parameter are. For example, it doesn't know that a reasonable Einstein radius is between
#    0.0" and 4.0"). How does it know what are reasonable values of parameters to guess?



import numpy as np
from scipy.integrate import quad
from scipy.optimize import root_scalar
from astropy import cosmology as cosmo
from skimage import measure


import autofit as af
from autolens import dimensions as dim
from autolens import text_util
from autolens.model.profiles import geometry_profiles

from autolens.array.grids import reshape_array_from_grid, reshape_returned_grid


class MassProfile(object):
    def convergence_func(self, eta):
        raise NotImplementedError("surface_density_func should be overridden")

    def convergence_from_grid(self, grid, return_in_2d=True, return_binned=True):
        pass
        # raise NotImplementedError("surface_density_from_grid should be overridden")

    def potential_from_grid(self, grid, return_in_2d=True, return_binned=True):
        pass
        # raise NotImplementedError("potential_from_grid should be overridden")

    def deflections_from_grid(self, grid, return_in_2d=True, return_binned=True):
        raise NotImplementedError("deflections_from_grid should be overridden")

    def mass_within_circle_in_units(
        self,
        radius: dim.Length,
        redshift_profile=None,
        redshift_source=None,
        unit_mass="solMass",
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        raise NotImplementedError()

    def mass_within_ellipse_in_units(
        self,
        major_axis: dim.Length,
        redshift_profile=None,
        redshift_source=None,
        unit_mass="solMass",
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        raise NotImplementedError()

    def einstein_radius_in_units(
        self, unit_length="arcsec", redshift_profile=None, cosmology=cosmo.Planck15
    ):
        return NotImplementedError()

    def einstein_mass_in_units(
        self,
        unit_mass="solMass",
        redshift_profile=None,
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        return NotImplementedError()

    def summarize_in_units(
        self,
        radii,
        prefix="",
        unit_length="arcsec",
        unit_mass="solMass",
        redshift_profile=None,
        redshift_source=None,
        cosmology=cosmo.Planck15,
        whitespace=80,
        **kwargs
    ):
        return ["Mass Profile = {}\n".format(self.__class__.__name__)]

    @property
    def unit_mass(self):
        return NotImplementedError()


# noinspection PyAbstractClass
class EllipticalMassProfile(geometry_profiles.EllipticalProfile, MassProfile):
    @af.map_types
    def __init__(
        self,
        centre: dim.Position = (0.0, 0.0),
        axis_ratio: float = 1.0,
        phi: float = 0.0,
    ):
        """
        Abstract class for elliptical mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        axis_ratio : float
            Ellipse's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of profile's ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalMassProfile, self).__init__(
            centre=centre, axis_ratio=axis_ratio, phi=phi
        )
        self.axis_ratio = axis_ratio
        self.phi = phi

    @dim.convert_units_to_input_units
    def mass_within_circle_in_units(
        self,
        radius: dim.Length,
        unit_mass="solMass",
        redshift_profile=None,
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """ Integrate the mass profiles's convergence profile to compute the total mass within a circle of \
        specified radius. This is centred on the mass profile.

        The following units for mass can be specified and output:

        - Dimensionless angular units (default) - 'solMass'.
        - Solar masses - 'solMass' (multiplies the angular mass by the critical surface mass density).

        Parameters
        ----------
        radius : dim.Length
            The radius of the circle to compute the dimensionless mass within.
        unit_mass : str
            The units the mass is returned in (angular | angular).
        critical_surface_density : float or None
            The critical surface mass density of the strong lens configuration, which converts mass from angulalr \
            units to phsical units (e.g. solar masses).
        """

        critical_surface_density = (
            kwargs["critical_surface_density"]
            if "critical_surface_density" in kwargs
            else None
        )

        mass = dim.Mass(
            value=quad(self.mass_integral, a=0.0, b=radius, args=(1.0,))[0],
            unit_mass=self.unit_mass,
        )

        return mass.convert(
            unit_mass=unit_mass, critical_surface_density=critical_surface_density
        )

    @dim.convert_units_to_input_units
    def mass_within_ellipse_in_units(
        self,
        major_axis: dim.Length,
        unit_mass="solMass",
        redshift_profile=None,
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """ Integrate the mass profiles's convergence profile to compute the total angular mass within an ellipse of \
        specified major axis. This is centred on the mass profile.

        The following units for mass can be specified and output:

        - Dimensionless angular units (default) - 'solMass'.
        - Solar masses - 'solMass' (multiplies the angular mass by the critical surface mass density)

        Parameters
        ----------
        major_axis : float
            The major-axis radius of the ellipse.
        unit_mass : str
            The units the mass is returned in (angular | angular).
        critical_surface_density : float or None
            The critical surface mass density of the strong lens configuration, which converts mass from angular \
            units to phsical units (e.g. solar masses).
        """

        critical_surface_density = (
            kwargs["critical_surface_density"]
            if "critical_surface_density" in kwargs
            else None
        )

        mass = dim.Mass(
            value=quad(
                self.mass_integral, a=0.0, b=major_axis, args=(self.axis_ratio,)
            )[0],
            unit_mass=self.unit_mass,
        )

        return mass.convert(
            unit_mass=unit_mass, critical_surface_density=critical_surface_density
        )

    def mass_integral(self, x, axis_ratio):
        """Routine to integrate an elliptical light profiles - set axis ratio to 1 to compute the luminosity within a \
        circle"""
        r = x * axis_ratio
        return 2 * np.pi * r * self.convergence_func(x)

    def density_between_circular_annuli_in_angular_units(
        self,
        inner_annuli_radius: dim.Length,
        outer_annuli_radius: dim.Length,
        unit_length="arcsec",
        unit_mass="solMass",
        redshift_profile=None,
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """Calculate the mass between two circular annuli and compute the density by dividing by the annuli surface
        area.

        The value returned by the mass integral is dimensionless, therefore the density between annuli is returned in \
        units of inverse radius squared. A conversion factor can be specified to convert this to a physical value \
        (e.g. the critical surface mass density).

        Parameters
        -----------
        inner_annuli_radius : float
            The radius of the inner annulus outside of which the density are estimated.
        outer_annuli_radius : float
            The radius of the outer annulus inside of which the density is estimated.
        """
        annuli_area = (np.pi * outer_annuli_radius ** 2.0) - (
            np.pi * inner_annuli_radius ** 2.0
        )

        outer_mass = self.mass_within_circle_in_units(
            radius=outer_annuli_radius,
            redshift_profile=redshift_profile,
            redshift_source=redshift_source,
            unit_mass=unit_mass,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        inner_mass = self.mass_within_circle_in_units(
            radius=inner_annuli_radius,
            redshift_profile=redshift_profile,
            redshift_source=redshift_source,
            unit_mass=unit_mass,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        return dim.MassOverLength2(
            value=(outer_mass - inner_mass) / annuli_area,
            unit_length=unit_length,
            unit_mass=unit_mass,
        )

    @dim.convert_units_to_input_units
    def average_convergence_of_1_radius_in_units(
        self,
        unit_length="arcsec",
        redshift_profile=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        """The radius a critical curve forms for this mass profile, e.g. where the mean convergence is equal to 1.0.

         In case of ellipitical mass profiles, the 'average' critical curve is used, whereby the convergence is \
         rescaled into a circle using the axis ratio.

         This radius corresponds to the Einstein radius of the mass profile, and is a property of a number of \
         mass profiles below.
         """

        kpc_per_arcsec = (
            kwargs["kpc_per_arcsec"] if "kpc_per_arcsec" in kwargs else None
        )

        def func(radius, redshift_profile, cosmology):
            radius = dim.Length(radius, unit_length=unit_length)
            return (
                self.mass_within_circle_in_units(
                    unit_mass="angular",
                    radius=radius,
                    redshift_profile=redshift_profile,
                    cosmology=cosmology,
                )
                - np.pi * radius ** 2.0
            )

        radius = (
            self.ellipticity_rescale
            * root_scalar(
                func, bracket=[1e-4, 1000.0], args=(redshift_profile, cosmology)
            ).root
        )
        radius = dim.Length(radius, unit_length)
        return radius.convert(unit_length=unit_length, kpc_per_arcsec=kpc_per_arcsec)

    @dim.convert_units_to_input_units
    def einstein_radius_in_units(
        self,
        unit_length="arcsec",
        redshift_profile=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        einstein_radius = self.average_convergence_of_1_radius_in_units(
            unit_length=unit_length,
            redshift_profile=redshift_profile,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        return dim.Length(einstein_radius, unit_length)

    @dim.convert_units_to_input_units
    def einstein_mass_in_units(
        self,
        unit_mass="solMass",
        redshift_profile=None,
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        einstein_radius = self.einstein_radius_in_units(
            unit_length=self.unit_length,
            redshift_profile=redshift_profile,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        return self.mass_within_circle_in_units(
            radius=einstein_radius,
            unit_mass=unit_mass,
            redshift_profile=redshift_profile,
            redshift_source=redshift_source,
            cosmology=cosmology,
            kwargs=kwargs,
        )

    @reshape_returned_grid
    def deflections_via_potential_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        potential_2d = self.potential_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        deflections_y_2d = np.gradient(potential_2d, grid.in_2d[:, 0, 0], axis=0)
        deflections_x_2d = np.gradient(potential_2d, grid.in_2d[0, :, 1], axis=1)

        return np.stack((deflections_y_2d, deflections_x_2d), axis=-1)

    @reshape_array_from_grid
    def lensing_jacobian_a11_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1.0 - np.gradient(deflections_2d[:, :, 1], grid.in_2d[0, :, 1], axis=1)

    @reshape_array_from_grid
    def lensing_jacobian_a12_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(deflections_2d[:, :, 1], grid.in_2d[:, 0, 0], axis=0)

    @reshape_array_from_grid
    def lensing_jacobian_a21_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return -1.0 * np.gradient(deflections_2d[:, :, 0], grid.in_2d[0, :, 1], axis=1)

    @reshape_array_from_grid
    def lensing_jacobian_a22_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        deflections_2d = self.deflections_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        return 1 - np.gradient(deflections_2d[:, :, 0], grid.in_2d[:, 0, 0], axis=0)

    def lensing_jacobian_from_grid(self, grid, return_in_2d=True, return_binned=True):

        a11 = self.lensing_jacobian_a11_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a12 = self.lensing_jacobian_a12_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a21 = self.lensing_jacobian_a21_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        a22 = self.lensing_jacobian_a22_from_grid(
            grid=grid, return_in_2d=return_in_2d, return_binned=return_binned
        )

        return np.array([[a11, a12], [a21, a22]])

    @reshape_array_from_grid
    def convergence_via_jacobian_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        convergence = 1 - 0.5 * (jacobian[0, 0] + jacobian[1, 1])

        return convergence

    @reshape_array_from_grid
    def shear_via_jacobian_from_grid(self, grid, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        gamma_1 = 0.5 * (jacobian[1, 1] - jacobian[0, 0])
        gamma_2 = -0.5 * (jacobian[0, 1] + jacobian[1, 0])

        return (gamma_1 ** 2 + gamma_2 ** 2) ** 0.5

    @reshape_array_from_grid
    def tangential_eigen_value_from_grid(
        self, grid, return_in_2d=True, return_binned=True
    ):

        convergence = self.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        shear = self.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        return 1 - convergence - shear

    @reshape_array_from_grid
    def radial_eigen_value_from_grid(self, grid, return_in_2d=True, return_binned=True):

        convergence = self.convergence_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        shear = self.shear_via_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        return 1 - convergence + shear

    @reshape_array_from_grid
    def magnification_from_grid(self, grid, return_in_2d=True, return_binned=True):

        jacobian = self.lensing_jacobian_from_grid(
            grid=grid, return_in_2d=False, return_binned=False
        )

        det_jacobian = jacobian[0, 0] * jacobian[1, 1] - jacobian[0, 1] * jacobian[1, 0]

        return 1 / det_jacobian

    def tangential_critical_curve_from_grid(self, grid):

        lambda_tangential_2d = self.tangential_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        tangential_critical_curve_indices = measure.find_contours(
            lambda_tangential_2d, 0
        )

        if tangential_critical_curve_indices == []:
            return []

        return grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=tangential_critical_curve_indices[0],
            shape=lambda_tangential_2d.shape,
        )

    def radial_critical_curve_from_grid(self, grid):

        lambda_radial_2d = self.radial_eigen_value_from_grid(
            grid=grid, return_in_2d=True, return_binned=False
        )

        radial_critical_curve_indices = measure.find_contours(lambda_radial_2d, 0)

        if radial_critical_curve_indices == []:
            return []

        return grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=radial_critical_curve_indices[0], shape=lambda_radial_2d.shape
        )

    def tangential_caustic_from_grid(self, grid):

        tangential_critical_curve = self.tangential_critical_curve_from_grid(grid=grid)

        if tangential_critical_curve == []:
            return []

        deflections_1d = self.deflections_from_grid(
            grid=tangential_critical_curve, return_in_2d=False, return_binned=False
        )

        return tangential_critical_curve - deflections_1d

    def radial_caustic_from_grid(self, grid):

        radial_critical_curve = self.radial_critical_curve_from_grid(grid=grid)

        if radial_critical_curve == []:
            return []

        deflections_1d = self.deflections_from_grid(
            grid=radial_critical_curve, return_in_2d=False, return_binned=False
        )

        return radial_critical_curve - deflections_1d

    def critical_curves_from_grid(self, grid):
        return [
            self.tangential_critical_curve_from_grid(grid=grid),
            self.radial_critical_curve_from_grid(grid=grid),
        ]

    def caustics_from_grid(self, grid):
        return [
            self.tangential_caustic_from_grid(grid=grid),
            self.radial_caustic_from_grid(grid=grid),
        ]

    @dim.convert_units_to_input_units
    def summarize_in_units(
        self,
        radii,
        prefix="",
        whitespace=80,
        unit_length="arcsec",
        unit_mass="solMass",
        redshift_profile=None,
        redshift_source=None,
        cosmology=cosmo.Planck15,
        **kwargs
    ):
        summary = super().summarize_in_units(
            radii=radii,
            prefix="",
            unit_length=unit_length,
            unit_mass=unit_mass,
            redshift_profile=redshift_profile,
            redshift_source=redshift_source,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        einstein_radius = self.einstein_radius_in_units(
            unit_length=unit_length,
            redshift_profile=redshift_profile,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        summary += [
            af.text_util.label_value_and_unit_string(
                label=prefix + "einstein_radius",
                value=einstein_radius,
                unit=unit_length,
                whitespace=whitespace,
            )
        ]

        einstein_mass = self.einstein_mass_in_units(
            unit_mass=unit_mass,
            redshift_profile=redshift_profile,
            redshift_source=redshift_source,
            cosmology=cosmology,
            kwargs=kwargs,
        )

        summary += [
            af.text_util.label_value_and_unit_string(
                label=prefix + "einstein_mass",
                value=einstein_mass,
                unit=unit_mass,
                whitespace=whitespace,
            )
        ]

        for radius in radii:
            mass = self.mass_within_circle_in_units(
                unit_mass=unit_mass,
                radius=radius,
                redshift_profile=redshift_profile,
                redshift_source=redshift_source,
                cosmology=cosmology,
                kwargs=kwargs,
            )

            summary += [
                text_util.within_radius_label_value_and_unit_string(
                    prefix=prefix + "mass",
                    radius=radius,
                    unit_length=unit_length,
                    value=mass,
                    unit_value=unit_mass,
                    whitespace=whitespace,
                )
            ]

        return summary

    @property
    def ellipticity_rescale(self):
        return NotImplementedError()



import autolens as al
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.model.inversion.plotters import inversion_plotters

# In this tutorial we'll use a new pixelization, called the VoronoiMagnification pixelization. This pixelization
# doesn't use a uniform grid of rectangular pixels, but instead uses an irregular 'Voronoi' pixels.
# So, why do we want to do that? Lets take another quick look at the rectangular grid..

# This simulates the same the image we've fitted in the past few tutorials.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(150, 150), pixel_scale=0.05, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.6
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.7,
            phi=135.0,
            intensity=0.2,
            effective_radius=0.2,
            sersic_index=2.5,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=1.0,
        add_noise=True,
    )


# Lets quickly remind ourselves of the image and the 3.0" circular mask we'll use to mask it.
ccd_data = simulate()

mask = al.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=2.5
)

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data, mask=mask)

# The lines of code below do everything we're used to, that is, setup an image, mask it, trace it via a tracer,
# setup the rectangular mapper, etc.
lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=1.6
    ),
)

lens_data = al.LensData(ccd_data=ccd_data, mask=mask)

source_galaxy = al.Galaxy(
    redshift=1.0,
    pixelization=pix.Rectangular(shape=(40, 40)),
    regularization=reg.Constant(coefficient=0.5),
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

fit = al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

al.lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)

# Okay, so lets think about the rectangular pixelization. Is this the optimal way to reconstruct our source? Are there
# features in the source-plane that arn't ideal? How do you think we could do a better job?

# Well, given we're doing a whole tutorial on using a different pixelization to the rectangular grid, you've probably
# guessed that it isn't optimal. Infact, its pretty rubbish, and not a pixelization we should actually want to model
# many lenses with!

# So what is wrong with the grid? Well, lets think about the source reconstruction.
inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_centres=True
)

# There is one clear problem, we are using a small number of the total source pixels to reconstruct the source. The
# majority of source pixels are located away from the source. By my estimate, we're using just 16 pixels (the
# central 4x4 grid) out of the 1600 pixels to actually fit the data! The remaining ~1500 pixels are doing *nothing*
# but fit noise.

# This is a waste and our analysis will take longer to run because of it. However, more importantly, it means that our
# Bayesian regularization scheme is sub-optimal. In tutorial 4, we discussed how the Bayesian evidence of the
# regularization wants to obtain the *simplest* source solution possible. That is the solution which fits the data well
# using the fewest source pixels. Clearly, if we dedicating a huge number of source pixels to doing *nothing*, our source
# reconstruction will be unecessarily complex (and therefore lower evidence).

# If our pixelization could 'focus' its pixels where we actually have more data, e.g. the highly magnified regions
# of the source-plane, we could reconstruct the source using far fewer pixels. That'd be great both for computational
# efficiency and increasing the Bayesian evidence and that is exactly what our Voronoi grid does.

# To achieve this, we first compute an 'image-plane pixelization grid', which is a set of sparse coordinates in the
# image-plane that will be ray-traced to the source-plane and define the centres of our source-pixel grid. We compute
# this grid directly from a pixelization, by passing it a grid.

adaptive = pix.VoronoiMagnification(shape=(20, 20))

pixelization_grid = adaptive.pixelization_grid_from_grid(grid=lens_data.grid)

# We can plot this grid over the image, to see that it is a coarse grid over-laying the image itself.

al.ccd_plotters.plot_image(
    ccd_data=ccd_data,
    grid=pixelization_grid,
    mask=mask,
    zoom_around_mask=True,
    extract_array_from_mask=True,
)

# When we pass a tracer a source galaxy with this pixelization it automatically computes the ray-traced source-plane
# Voronoi grid using the grid above. Thus, our Voronoi pixelization is used by the tracer's fit.

source_galaxy = al.Galaxy(
    redshift=1.0, pixelization=adaptive, regularization=reg.Constant(coefficient=1.0)
)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

# If we look at the lens fit, we'll that our source-plane no longer uses rectangular pixels, but Voronoi pixels!

fit = al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

al.lens_fit_plotters.plot_fit_subplot(
    fit=fit,
    should_plot_image_plane_pix=True,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

# And we can take a closer inspection of the inversion itself.
inversion_plotters.plot_pixelization_values(
    inversion=fit.inversion, should_plot_centres=True
)

# Clearly, this is an improvement. We're using fewer pixels than the rectangular grid (400, instead of 1600), but
# reconstructing our source is far greater detail. A win all around? It sure is.

# On our rectangular grid, we regularized each source pixel with its 4 neighbors. We compared their fluxes, summed the
# differences, and penalized solutions where the differences were large. For a Voronoi grid, we do the same thing,
# now comparing each source-pixel with all other source-pixels with which it shares a direct vertex. This means that
# different source-pixels may be regularized with different numbers of source-pixels, depending on how many neighbors
# are formed.

# This Voronoi magnification grid is still far from optimal. There are lots of source-pixels effectively fitting just
# noise. We may achieve even better solutions if the central regions of the source were reconstructed using even more
# pixels. So, how do we improve on this? Well, you'll have to wait until chapter 5, when we introduce PyAutoLens's
# adaptive functionality, or 'hyper-mode'.

# In the mean time, you may wish to experiment with using both Rectangular and VoronoiMagnification grids to fit
# lenses which can be easily achieve by changing the input pixeliation given to a pipeline.



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
import itertools
from scipy.spatial import Voronoi

from autolens.plotters import plotter_util, grid_plotters
from autolens.model.inversion import mappers
from autolens.data.plotters import ccd_plotters


def plot_image_and_mapper(
    ccd_data,
    mapper,
    mask=None,
    positions=None,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    units="arcsec",
    kpc_per_arcsec=None,
    output_path=None,
    output_filename="image_and_mapper",
    output_format="show",
):

    rows, columns, figsize = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=2
    )
    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    ccd_plotters.plot_image(
        ccd_data=ccd_data,
        mask=mask,
        positions=positions,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=None,
        xyticksize=16,
        norm="linear",
        norm_min=None,
        norm_max=None,
        linthresh=0.05,
        linscale=0.01,
        figsize=None,
        aspect="square",
        cmap="jet",
        cb_ticksize=10,
        titlesize=10,
        xlabelsize=10,
        ylabelsize=10,
        output_path=output_path,
        output_format=output_format,
    )

    image_grid = convert_grid(
        grid=mapper.grid.unlensed_unsubbed_1d,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
    )

    point_colors = itertools.cycle(["y", "r", "k", "g", "m"])
    plot_image_plane_image_pixels(
        grid=image_grid, image_pixels=image_pixels, point_colors=point_colors
    )
    plot_image_plane_source_pixels(
        grid=image_grid,
        mapper=mapper,
        source_pixels=source_pixels,
        point_colors=point_colors,
    )

    plt.subplot(rows, columns, 2)

    plot_mapper(
        mapper=mapper,
        should_plot_centres=should_plot_centres,
        should_plot_grid=should_plot_grid,
        should_plot_border=should_plot_border,
        image_pixels=image_pixels,
        source_pixels=source_pixels,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=None,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def plot_mapper(
    mapper,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    xyticksize=16,
    figsize=(7, 7),
    title="Mapper",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    output_path=None,
    output_filename="mapper",
    output_format="show",
):

    if isinstance(mapper, mappers.RectangularMapper):

        plot_rectangular_mapper(
            mapper=mapper,
            should_plot_centres=should_plot_centres,
            should_plot_grid=should_plot_grid,
            should_plot_border=should_plot_border,
            image_pixels=image_pixels,
            source_pixels=source_pixels,
            as_subplot=as_subplot,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            xyticksize=xyticksize,
            figsize=figsize,
            title=title,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            output_path=output_path,
            output_filename=output_filename,
            output_format=output_format,
        )


def plot_rectangular_mapper(
    mapper,
    should_plot_centres=False,
    should_plot_grid=False,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    xyticksize=16,
    figsize=(7, 7),
    title="Rectangular Mapper",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    output_path=None,
    output_filename="rectangular_mapper",
    output_format="show",
):

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    set_axis_limits(mapper=mapper, units=units, kpc_per_arcsec=kpc_per_arcsec)
    plot_rectangular_pixelization_lines(
        mapper=mapper, units=units, kpc_per_arcsec=kpc_per_arcsec
    )

    plotter_util.set_title(title=title, titlesize=titlesize)
    grid_plotters.set_xy_labels(
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
    )

    plot_centres(
        should_plot_centres=should_plot_centres,
        mapper=mapper,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
    )

    plot_plane_grid(
        should_plot_grid=should_plot_grid,
        mapper=mapper,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        pointsize=10,
        xyticksize=xyticksize,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
    )

    plot_border(
        should_plot_border=should_plot_border,
        mapper=mapper,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        pointsize=30,
        xyticksize=xyticksize,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
    )

    mapper_grid = convert_grid(
        grid=mapper.grid, units=units, kpc_per_arcsec=kpc_per_arcsec
    )

    point_colors = itertools.cycle(["y", "r", "k", "g", "m"])
    plot_source_plane_image_pixels(
        grid=mapper_grid, image_pixels=image_pixels, point_colors=point_colors
    )
    plot_source_plane_source_pixels(
        grid=mapper_grid,
        mapper=mapper,
        source_pixels=source_pixels,
        point_colors=point_colors,
    )

    plotter_util.output_figure(
        None,
        as_subplot=as_subplot,
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plotter_util.close_figure(as_subplot=as_subplot)


def plot_voronoi_mapper(
    mapper,
    source_pixel_values,
    should_plot_centres=True,
    should_plot_grid=True,
    should_plot_border=False,
    image_pixels=None,
    source_pixels=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    xyticksize=16,
    figsize=(7, 7),
    title="Rectangular Mapper",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    output_path=None,
    output_filename="voronoi_mapper",
    output_format="show",
):

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    set_axis_limits(mapper=mapper, units=units, kpc_per_arcsec=kpc_per_arcsec)

    regions_SP, vertices_SP = voronoi_finite_polygons_2d(mapper.voronoi)

    color_values = source_pixel_values[:] / np.max(source_pixel_values)
    cmap = plt.get_cmap("jet")

    set_colorbar(
        cmap=cmap,
        color_values=source_pixel_values,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
    )

    for region, index in zip(regions_SP, range(mapper.pixels)):
        polygon = vertices_SP[region]
        col = cmap(color_values[index])
        plt.fill(*zip(*polygon), alpha=0.7, facecolor=col, lw=0.0)

    plotter_util.set_title(title=title, titlesize=titlesize)
    grid_plotters.set_xy_labels(
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
    )

    plot_centres(
        should_plot_centres=should_plot_centres,
        mapper=mapper,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
    )

    plot_plane_grid(
        should_plot_grid=should_plot_grid,
        mapper=mapper,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        pointsize=10,
        xyticksize=xyticksize,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
    )

    plot_border(
        should_plot_border=should_plot_border,
        mapper=mapper,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        pointsize=30,
        xyticksize=xyticksize,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
    )

    mapper_grid = convert_grid(
        grid=mapper.grid, units=units, kpc_per_arcsec=kpc_per_arcsec
    )

    point_colors = itertools.cycle(["y", "r", "k", "g", "m"])
    plot_source_plane_image_pixels(
        grid=mapper_grid, image_pixels=image_pixels, point_colors=point_colors
    )
    plot_source_plane_source_pixels(
        grid=mapper_grid,
        mapper=mapper,
        source_pixels=source_pixels,
        point_colors=point_colors,
    )

    plotter_util.output_figure(
        None,
        as_subplot=as_subplot,
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plotter_util.close_figure(as_subplot=as_subplot)


def voronoi_finite_polygons_2d(vor, radius=None):
    """
    Reconstruct infinite voronoi regions in a 2D diagram to finite
    regions.
    Parameters
    ----------
    vor : Voronoi
        Input diagram
    radius : float, optional
        Distance to 'points at infinity'.
    Returns
    -------
    regions : list of tuples
        Indices of vertices in each revised Voronoi regions.
    vertices : list of tuples
        Coordinates for revised Voronoi vertices. Same as coordinates
        of input vertices, with 'points at infinity' appended to the
        end.
    """

    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max() * 2

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge

            t = vor.points[p2] - vor.points[p1]  # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # hyper

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)


def plot_rectangular_pixelization_lines(mapper, units, kpc_per_arcsec):

    if units in "arcsec" or kpc_per_arcsec is None:

        ys = np.linspace(
            mapper.geometry.arc_second_minima[0],
            mapper.geometry.arc_second_maxima[0],
            mapper.shape[0] + 1,
        )
        xs = np.linspace(
            mapper.geometry.arc_second_minima[1],
            mapper.geometry.arc_second_maxima[1],
            mapper.shape[1] + 1,
        )

    elif units in "kpc":

        ys = np.linspace(
            mapper.geometry.arc_second_minima[0] * kpc_per_arcsec,
            mapper.geometry.arc_second_maxima[0] * kpc_per_arcsec,
            mapper.shape[0] + 1,
        )
        xs = np.linspace(
            mapper.geometry.arc_second_minima[1] * kpc_per_arcsec,
            mapper.geometry.arc_second_maxima[1] * kpc_per_arcsec,
            mapper.shape[1] + 1,
        )

    # grid lines
    for x in xs:
        plt.plot([x, x], [ys[0], ys[-1]], color="black", linestyle="-")
    for y in ys:
        plt.plot([xs[0], xs[-1]], [y, y], color="black", linestyle="-")


def set_axis_limits(mapper, units, kpc_per_arcsec):

    if units in "arcsec" or kpc_per_arcsec is None:

        grid_plotters.set_axis_limits(
            axis_limits=np.asarray(
                [
                    mapper.geometry.arc_second_minima[1],
                    mapper.geometry.arc_second_maxima[1],
                    mapper.geometry.arc_second_minima[0],
                    mapper.geometry.arc_second_maxima[0],
                ]
            ),
            grid=None,
            symmetric_around_centre=False,
        )

    elif units in "kpc":

        grid_plotters.set_axis_limits(
            axis_limits=np.asarray(
                [
                    mapper.geometry.arc_second_minima[1] * kpc_per_arcsec,
                    mapper.geometry.arc_second_maxima[1] * kpc_per_arcsec,
                    mapper.geometry.arc_second_minima[0] * kpc_per_arcsec,
                    mapper.geometry.arc_second_maxima[0] * kpc_per_arcsec,
                ]
            ),
            grid=None,
            symmetric_around_centre=False,
        )


def set_colorbar(
    cmap, color_values, cb_fraction, cb_pad, cb_tick_values, cb_tick_labels
):

    cax = cm.ScalarMappable(cmap=cmap)
    cax.set_array(color_values)

    if cb_tick_values is None and cb_tick_labels is None:
        plt.colorbar(mappable=cax, fraction=cb_fraction, pad=cb_pad)
    elif cb_tick_values is not None and cb_tick_labels is not None:
        cb = plt.colorbar(
            mappable=cax, fraction=cb_fraction, pad=cb_pad, ticks=cb_tick_values
        )
        cb.ax.set_yticklabels(cb_tick_labels)


def plot_centres(should_plot_centres, mapper, units, kpc_per_arcsec):

    if should_plot_centres:

        if units in "arcsec" or kpc_per_arcsec is None:

            pixel_centres = mapper.geometry.pixel_centres

        elif units in "kpc":

            pixel_centres = mapper.geometry.pixel_centres * kpc_per_arcsec

        plt.scatter(y=pixel_centres[:, 0], x=pixel_centres[:, 1], s=3, c="r")


def plot_plane_grid(
    should_plot_grid,
    mapper,
    as_subplot,
    units,
    kpc_per_arcsec,
    pointsize,
    xyticksize,
    title,
    titlesize,
    xlabelsize,
    ylabelsize,
):

    if should_plot_grid:

        grid_units = convert_grid(
            grid=mapper.grid, units=units, kpc_per_arcsec=kpc_per_arcsec
        )

        grid_plotters.plot_grid(
            grid=grid_units,
            as_subplot=as_subplot,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            pointsize=pointsize,
            xyticksize=xyticksize,
            title=title,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
        )


def plot_border(
    should_plot_border,
    mapper,
    as_subplot,
    units,
    kpc_per_arcsec,
    pointsize,
    xyticksize,
    title,
    titlesize,
    xlabelsize,
    ylabelsize,
):

    if should_plot_border:

        border_arcsec = mapper.grid[mapper.grid.sub_border_pixels]
        border_units = convert_grid(
            grid=border_arcsec, units=units, kpc_per_arcsec=kpc_per_arcsec
        )

        grid_plotters.plot_grid(
            grid=border_units,
            as_subplot=as_subplot,
            units=units,
            kpc_per_arcsec=kpc_per_arcsec,
            pointsize=pointsize,
            pointcolor="y",
            xyticksize=xyticksize,
            title=title,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
        )


def plot_image_plane_image_pixels(grid, image_pixels, point_colors):

    if image_pixels is not None:

        for image_pixel_set in image_pixels:
            color = next(point_colors)
            plt.scatter(
                y=np.asarray(grid[image_pixel_set, 0]),
                x=np.asarray(grid[image_pixel_set, 1]),
                color=color,
                s=10.0,
            )


def plot_image_plane_source_pixels(grid, mapper, source_pixels, point_colors):

    if source_pixels is not None:

        for source_pixel_set in source_pixels:
            color = next(point_colors)
            for source_pixel in source_pixel_set:
                plt.scatter(
                    y=np.asarray(
                        grid[
                            mapper.sub_mask_1d_index_to_pixelization_1d_index[
                                source_pixel
                            ],
                            0,
                        ]
                    ),
                    x=np.asarray(
                        grid[
                            mapper.sub_mask_1d_index_to_pixelization_1d_index[
                                source_pixel
                            ],
                            1,
                        ]
                    ),
                    s=8,
                    color=color,
                )


def plot_source_plane_image_pixels(grid, image_pixels, point_colors):

    if image_pixels is not None:

        for image_pixel_set in image_pixels:
            color = next(point_colors)
            plt.scatter(
                y=np.asarray(grid[[image_pixel_set], 0]),
                x=np.asarray(grid[[image_pixel_set], 1]),
                s=8,
                color=color,
            )


def plot_source_plane_source_pixels(grid, mapper, source_pixels, point_colors):

    if source_pixels is not None:

        for source_pixel_set in source_pixels:
            color = next(point_colors)
            for source_pixel in source_pixel_set:
                plt.scatter(
                    y=np.asarray(
                        grid[
                            mapper.sub_mask_1d_index_to_pixelization_1d_index[
                                source_pixel
                            ],
                            0,
                        ]
                    ),
                    x=np.asarray(
                        grid[
                            mapper.sub_mask_1d_index_to_pixelization_1d_index[
                                source_pixel
                            ],
                            1,
                        ]
                    ),
                    s=8,
                    color=color,
                )


def convert_grid(grid, units, kpc_per_arcsec):

    if units in "arcsec" or kpc_per_arcsec is None:
        return grid
    elif units in "kpc":
        return grid * kpc_per_arcsec



import autofit as af
import autolens as al
from autolens.pipeline.phase import phase_imaging
from autolens.model.galaxy import galaxy_model as gm

# We finished the last tutorial on a sour note. Our non-linear search failed miserably, and we were unable to infer a
# lens model which fitted our data. In this tutorial, we're going to right our past wrongs and infer
# the correct model - not just once, but three times!

# You need to change the path below to the chapter 1 directory.
chapter_path = "/path/to/user/autolens_workspace/howtolens/chapter_2_lens_modeling/"
chapter_path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/howtolens/chapter_2_lens_modeling/"

af.conf.instance = af.conf.Config(
    config_path=chapter_path + "configs/t4_dealing_with_failure",
    output_path=chapter_path + "output",
)

# Even with my custom config files the non-linear searches will take a bit of time to run in this tutorial if you choose
# to run them yourselves.

# Another simulate image function, albeit it generates a new image.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(130, 130), pixel_scale=0.1, sub_grid_size=2
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.9,
            phi=45.0,
            intensity=0.04,
            effective_radius=0.5,
            sersic_index=3.5,
        ),
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=0.8
        ),
    )

    source_galaxy = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.5,
            phi=90.0,
            intensity=0.03,
            effective_radius=0.3,
            sersic_index=3.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy, source_galaxy])

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.1,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Simulate the CCD data and plot it.
ccd_data = simulate()

al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

### Approach 1 -  Prior Tuning ###

# First, we're going to try giving our non-linear search a helping hand. Lets think about
# our priors, they're telling the non-linear search where to look in parameter space. If we tell it to
# look in the right place (that is, *tune* our priors), it might find the best-fit lens model.

# We've already seen that we can fully customize priors in PyAutoLens, so lets do it. I've set up a custom phase
# below and specified a new set of priors that'll give the non-linear search a much better chance if inferring the
# correct model. I've also let you know what we're changing the priors from (as initially specified by the
# 'config/priors/default' config files.)


class CustomPriorPhase(phase_imaging.PhaseImaging):
    def customize_priors(self, results):

        # We've called our lens galaxy 'lens' this time, for shorter more readable code.

        # By default, the prior on the (y,x) coordinates of a light / mass profile is a GaussianPrior with mean
        # 0.0" and sigma "1.0. However, visual inspection of our strong lens image tells us that its clearly around
        # x = 0.0" and y = 0.0", so lets reduce where non-linear search looks for these parameters.

        self.galaxies.lens.light.centre_0 = af.UniformPrior(
            lower_limit=-0.05, upper_limit=0.05
        )

        self.galaxies.lens.light.centre_1 = af.UniformPrior(
            lower_limit=-0.05, upper_limit=0.05
        )

        self.galaxies.lens.mass.centre_0 = af.UniformPrior(
            lower_limit=-0.05, upper_limit=0.05
        )

        self.galaxies.lens.mass.centre_1 = af.UniformPrior(
            lower_limit=-0.05, upper_limit=0.05
        )

        # By default, the axis-ratio of our lens galaxy's elliptical light profile is a UniformPrior between 0.2 and
        # 1.0. However, by looking at the image it looks fairly circular, so lets use a GaussianPrior nearer 1.0.
        self.galaxies.lens.light.axis_ratio = af.GaussianPrior(mean=0.8, sigma=0.15)

        # We'll also assume that the light profile's axis_ratio informs us of the mass-profile's axis_ratio, but
        # because this may not strictly be true (e.g. because of dark matter) we'll use a wider prior.
        self.galaxies.lens.mass.axis_ratio = af.GaussianPrior(mean=0.8, sigma=0.25)

        # By default, the orientation of the galaxy's light profile, phi, uses a UniformPrior between 0.0 and
        # 180.0 degrees. However, if you look really close at the image (and maybe adjust the color-map of the plot)
        # you'll notice that it is elliptical and oriented around 45.0 degrees counter-clockwise
        # from the x-axis. Lets update our prior
        self.galaxies.lens.light.phi = af.GaussianPrior(mean=45.0, sigma=15.0)

        # Again, lets assume that the light's orientation roughly traces that of the mass.
        self.galaxies.lens.mass.phi = af.GaussianPrior(mean=45.0, sigma=30.0)

        # The effective radius of a light profile is its 'half-light' radius, the radius at which 50% of its
        # total luminosity is internal to a circle defined within that radius. PyAutoLens assumes a
        # UniformPrior on this quantity between 0.0" and 4.0", but inspection of the image (again, using a colormap
        # scaling) shows the lens's light doesn't extend anywhere near 4.0", so lets reduce it.
        self.galaxies.lens.light.effective_radius = af.GaussianPrior(
            mean=0.5, sigma=0.8
        )

        # Typically, we have knowledge of our lens galaxy's morphology. Most strong lenses are massive elliptical
        # galaxies which have Sersic indexes near 4. So lets change our Sersic index from a UniformPrior between
        # 0.8 and 8.0 to reflect this.
        self.galaxies.lens.light.sersic_index = af.GaussianPrior(mean=4.0, sigma=1.0)

        # Finally, the 'ring' that the lensed source forms clearly has a radius of about 0.8". This is its Einstein
        # radius, so lets change the prior from a UniformPrior between 0.0" and 4.0".
        self.galaxies.lens.mass.einstein_radius = af.GaussianPrior(mean=0.8, sigma=0.2)

        # In this exercise, I'm not going to change any priors on the source galaxy. Whilst lens modeling experts can
        # look at a strong lens and often tell you roughly where the source-galaxy is located (in the source-plane),
        # it is something of art form. Furthermore, the source's morphology can be pretty complex, making it
        # difficult to come up with a good source prior!


# We can now create this custom phase and run it. Our non-linear search will start in a high likelihood region
# of parameter space.
custom_prior_phase = CustomPriorPhase(
    phase_name="phase_t4_tuned_priors",
    galaxies=dict(
        lens=gm.GalaxyModel(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic,
            mass=al.mass_profiles.EllipticalIsothermal,
        ),
        source=gm.GalaxyModel(
            redshift=1.0, light=al.light_profiles.EllipticalExponential
        ),
    ),
    optimizer_class=af.MultiNest,
)

print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t4_dealing_with_failure"
    "folder for live output of the results, images and lens model."
    "This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)

custom_prior_result = custom_prior_phase.run(data=ccd_data)

print("MultiNest has finished run - you may now continue the notebook.")

# Bam! We get a good model. The right model. A glorious model! We gave our non-linear search a helping hand, and it
# repaid us in spades!

# Check out the PDF in the '/howstolens/chapter_2_lens_modeling/output/4_custom_priors/image' folder - what degeneracies
# do you notice between parameters?
al.lens_fit_plotters.plot_fit_subplot(fit=custom_prior_result.most_likely_fit)

# Okay, so we've learnt that by tuning our priors to the lens we're fitting we can increase our chance of inferring the
# global maxima lens model. Before moving onto the next approach, lets think about the advantages and disadvantages of prior
# tuning:

# Advantage - We find the maximum likelihood solution in parameter space.
# Advantage - The phase took less time to run because the non-linear search explored less of parameter space.
# Disadvantage - If we specified one prior incorrectly the non-linear search would begin and therefore end at
#                an incorrect solution.
# Disadvantage - Our phase was tailored to this specific strong lens. If we want to fit a large sample of lenses we'd
#                have to write a custom phase for every single one - this would take up a lot of our time!

### Approach 2 -  Reducing Complexity ###

# Previously, Our non-linear searched failed because we made the lens model more complex. Maybe we
# can make it less complex, whilst still keeping it fairly realistic? Maybe there are some assumptions we can make
# to reduce the number of lens model parameters and therefore dimensionality of non-linear parameter space?

# Well, we can *always* make assumptions. Below, I'm going to create a phase that assumes that
# light-traces-mass. That is, that our light-profile's centre, axis_ratio and orientation are perfectly
# aligned with its mass. This may, or may not, be a reasonable assumption, but it'll remove 4 parameters from the lens
# model (the mass-profiles y, x, axis_ratio and phi), so its worth trying!


class LightTracesMassPhase(phase_imaging.PhaseImaging):
    def customize_priors(self, results):

        # In the pass priors function we can 'pair' any two parameters by setting them equal to one another. This
        # removes the parameter on the left-hand side of the pairing from the lens model such that is always assumes
        # the same value as the parameter on the right-hand side.
        self.galaxies.lens.mass.centre_0 = self.galaxies.lens.light.centre_0

        # Now, the mass-profile's y coordinate will only use the y coordinate of the light profile. Lets do this with
        # the remaining geometric parameters of the light and mass profiles
        self.galaxies.lens.mass.centre_1 = self.galaxies.lens.light.centre_1

        self.galaxies.lens.mass.axis_ratio = self.galaxies.lens.light.axis_ratio

        self.galaxies.lens.mass.phi = self.galaxies.lens.light.phi


# Again, we create this phase and run it. The non-linear search now has a less complex parameter space to search.
light_traces_mass_phase = LightTracesMassPhase(
    phase_name="phase_t4_light_traces_mass",
    galaxies=dict(
        lens=gm.GalaxyModel(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic,
            mass=al.mass_profiles.EllipticalIsothermal,
        ),
        source=gm.GalaxyModel(
            redshift=1.0, light=al.light_profiles.EllipticalExponential
        ),
    ),
    optimizer_class=af.MultiNest,
)

print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t4_dealing_with_failure"
    "folder for live output of the results, images and lens model."
    "This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)

light_traces_mass_phase_result = light_traces_mass_phase.run(data=ccd_data)

print("MultiNest has finished run - you may now continue the notebook.")
al.lens_fit_plotters.plot_fit_subplot(
    fit=light_traces_mass_phase_result.most_likely_fit
)

# The results look pretty good. Our source galaxy fits the data pretty well and we've clearly inferred a model that
# looks similar to the one above. However, inspection of the residuals shows that the fit wasn't quite as good as the
# custom-phase above.

# It turns out that when I simulated this image light didn't perfectly trace mass. The light-profile's axis-ratio was
# 0.9, whereas the mass-profiles was 0.8. The quality of the fit has suffered as a result and the likelihood we've
# inferred is lower.
#
# Herein lies the pitfalls of making assumptions - they may make your model less realistic and your results
# worse! Nevertheless, our lens model is clearly much better than it was in the previous tutorial, so making
# assumptions isn't a bad idea if you're struggling to fit the data well.

# Again, lets consider the advantages and disadvantages of this approach:

# Advantage - By reducing parameter space's complexity we inferred a global maximum likelihood.
# Advantage - The phase is not specific to one lens - we could run it on many strong lens images.
# Disadvantage - Our model was less realistic and our fit suffered as a result.

### Approach 3: Look Harder ###

# In approaches 1 and 2 we extended our non-linear search an olive branch and helped it find the highest
# likelihood regions of parameter space. In approach 3 ,we're going to tell it to just 'look harder'.

# Basically, every non-linear search algorithm has a set of parameters that govern how thoroughly it searches parameter
# space. The more thoroughly it looks, the more likely it is that it'll find the global maximum lens model. However,
# the search will also take longer - and we don't want it to take too long to get some results.

# Lets setup a phase and overwrite some of the non-linear search's parameters from the defaults it assumes in the
# 'config/non_linear.ini' config file:

custom_non_linear_phase = phase_imaging.PhaseImaging(
    phase_name="phase_t4_custom_non_linear",
    galaxies=dict(
        lens=gm.GalaxyModel(
            redshift=0.5,
            light=al.light_profiles.EllipticalSersic,
            mass=al.mass_profiles.EllipticalIsothermal,
        ),
        source=gm.GalaxyModel(
            redshift=1.0, light=al.light_profiles.EllipticalExponential
        ),
    ),
    optimizer_class=af.MultiNest,
)

# The 'optimizer' below is MultiNest, the non-linear search we're using.

# When MultiNest searches non-linear parameter space it places down a set of 'live-points', each of which corresponds
# to a particular lens model (set of parameters) with an associted likelihood. When it samples a new lens model with a
# higher likelihood than any of the currently active live points, this new lens model becomes an active point. As a
# result, the active point with the lowest likelihood is discarded.

# The more live points MultiNest uses the more thoroughly it will sample parameter space. Lets increase the number of
# points from the default value (50) to 100.
custom_non_linear_phase.optimizer.n_live_points = 100

# When MultiNest find a 'peak' likelihood in parameter space it begins to converge around this peak. It
# does this by guessing lens models with similar parameters. However, this peak might not be the global maximum,
# and if MultiNest converges too quickly around a peak it won't realise this before its too late.
#
# The sampling efficiency therefore describes how quickly MultiNest converges around a peak. It assumes values between
# 0.0 and 1.0, where 1.0 corresponds to the fastest convergence but highest risk of not locating the global maximum.
# Lets reduce the sampling efficiency from 0.8 to 0.5.
custom_non_linear_phase.optimizer.sampling_efficiency = 0.5

# These are the two most important MultiNest parameters controlling how it navigates parameter space, so lets run this
# phase and see if our more detailed inspection of parameter space finds the correct lens model.
print(
    "MultiNest has begun running - checkout the workspace/howtolens/chapter_2_lens_modeling/output/t4_dealing_with_failure"
    "folder for live output of the results, images and lens model."
    "This Jupyter notebook cell with progress once MultiNest has completed - this could take some time!"
)
custom_non_linear_result = custom_non_linear_phase.run(data=ccd_data)
print("MultiNest has finished run - you may now continue the notebook.")

al.lens_fit_plotters.plot_fit_subplot(fit=custom_non_linear_result.most_likely_fit)

# Indeed, it does. Thus, we can always brute-force our way to a good lens model, if all else fails.

# Finally, lets list the advantages and disadvantages of this approach:

# Advantage - Its easy to setup, we just increase n_live_points or decrease sampling_efficiency.
# Advantage - It generalizes to any strong lens.
# Advantage - We didn't have to make our model less realistic.
# Disadvantage - Its expensive. Very expensive. The run-time of this phase was over 6 hours. For more complex models
#                we could be talking days or weeks (or, dare I say it, months).

# So, there we have it, we can now fit strong lenses with PyAutoLens. And if it fails, we know how to get it to work.
# I hope you're feeling pretty smug. You might even be thinking 'why should I bother with the rest of these tutorials, if I
# can fit strong a lens already'.

# Well, my friend, I want you to think about the last disadvantage listed above. If modeling a single lens could
# really take as long as a month, are you really willing to spend your valuable time waiting for this? I'm not, which
# is why I developed PyAutoLens, and in the next tutorial we'll see how we can get the best of both worlds - realistic,
# complex lens model that take mere hours to infer!

# Before doing that though, I want you to go over the advantages and disadvantages listed above again and think
# whether we could combine these different approaches to get the best of all worlds.



import autolens as al
import os

import numpy as np


test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestMap1dIndexesTo2dIndex(object):
    def test__9_1d_indexes_from_0_to_8__map_to_shape_3x3(self):

        indexes_1d = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])

        indexes_2d = al.array_mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(3, 3)
        )

        assert (
            indexes_2d
            == np.array(
                [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
            )
        ).all()

    def test__6_1d_indexes_from_0_to_5__map_to_shape_2x3(self):

        indexes_1d = np.array([0, 1, 2, 3, 4, 5])

        indexes_2d = al.array_mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(2, 3)
        )

        assert (
            indexes_2d == np.array([[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]])
        ).all()

    def test__6_1d_indexes_from_0_to_5__map_to_shape_3x2(self):

        indexes_1d = np.array([0, 1, 2, 3, 4, 5])

        indexes_2d = al.array_mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(3, 2)
        )

        assert (
            indexes_2d == np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]])
        ).all()

    def test__9_1d_indexes_from_0_to_8_different_order__map_to_shape_3x3(self):

        indexes_1d = np.array([1, 4, 7, 8, 0, 2, 3, 5, 6])

        indexes_2d = al.array_mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(3, 3)
        )

        assert (
            indexes_2d
            == np.array(
                [[0, 1], [1, 1], [2, 1], [2, 2], [0, 0], [0, 2], [1, 0], [1, 2], [2, 0]]
            )
        ).all()


class TestMap2dIndexesTo1dIndex(object):
    def test__9_2d_indexes_from_0_0_to_2_2__map_to_shape_3x3(self):

        indexes_2d = np.array(
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
        )

        indexes_1d = al.array_mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(3, 3)
        )

        assert (indexes_1d == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()

    def test__6_1d_indexes_from_0_0_to_1_2__map_to_shape_2x3(self):

        indexes_2d = np.array([[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]])

        indexes_1d = al.array_mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(2, 3)
        )

        assert (indexes_1d == np.array([0, 1, 2, 3, 4, 5])).all()

    def test__6_1d_indexes_from_0_0_to_2_1__map_to_shape_3x2(self):

        indexes_2d = np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]])

        indexes_1d = al.array_mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(3, 2)
        )

        assert (indexes_1d == np.array([0, 1, 2, 3, 4, 5])).all()

    def test__9_1d_indexes_from_0_0_to_2_2_different_order__map_to_shape_3x3(self):

        indexes_2d = np.array(
            [[0, 1], [1, 1], [2, 1], [2, 2], [0, 0], [0, 2], [1, 0], [1, 2], [2, 0]]
        )

        indexes_1d = al.array_mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(3, 3)
        )

        assert (indexes_1d == np.array([1, 4, 7, 8, 0, 2, 3, 5, 6])).all()


class TestSubArray1DFromSubArray2d(object):
    def test__map_simple_data__sub_grid_size_1(self):

        array_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([5])).all()

        array_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        mask = np.array(
            [[True, False, True], [False, False, False], [True, False, True]]
        )

        array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([2, 4, 5, 6, 8])).all()

        array_2d = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])

        mask = np.array(
            [
                [True, False, True, True],
                [False, False, False, True],
                [True, False, True, False],
            ]
        )

        array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([2, 5, 6, 7, 10, 12])).all()

        array_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])

        mask = np.array(
            [
                [True, False, True],
                [False, False, False],
                [True, False, True],
                [True, True, True],
            ]
        )

        array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([2, 4, 5, 6, 8])).all()

    def test__map_simple_data__sub_grid_size_2(self):

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
            ]
        )

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        sub_array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (sub_array_1d == np.array([15, 16, 3, 4])).all()

        mask = np.array([[True, False, True], [True, False, True], [True, True, False]])

        sub_array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (
            sub_array_1d == np.array([3, 4, 9, 10, 15, 16, 3, 4, 11, 12, 17, 18])
        ).all()

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6, 7, 7, 7],
                [7, 8, 9, 10, 11, 12, 7, 7, 7],
                [13, 14, 15, 16, 17, 18, 7, 7, 7],
                [1, 2, 3, 4, 5, 6, 7, 7, 7],
                [7, 8, 9, 10, 11, 12, 7, 7, 7],
                [13, 14, 15, 16, 17, 18, 7, 7, 7],
            ]
        )

        mask = np.array(
            [
                [True, False, True, True],
                [False, False, False, True],
                [True, False, True, False],
            ]
        )

        sub_array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (
            sub_array_1d
            == np.array(
                [
                    3,
                    4,
                    9,
                    10,
                    13,
                    14,
                    1,
                    2,
                    15,
                    16,
                    3,
                    4,
                    17,
                    18,
                    5,
                    6,
                    9,
                    10,
                    15,
                    16,
                    7,
                    7,
                    7,
                    7,
                ]
            )
        ).all()

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [7, 7, 7, 7, 7, 7],
                [7, 7, 7, 7, 7, 7],
            ]
        )

        mask = np.array(
            [
                [True, False, True],
                [False, False, False],
                [True, False, True],
                [True, True, True],
            ]
        )

        sub_array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (
            sub_array_1d
            == np.array(
                [3, 4, 9, 10, 13, 14, 1, 2, 15, 16, 3, 4, 17, 18, 5, 6, 9, 10, 15, 16]
            )
        ).all()

    def test__setup_2x2_data__sub_grid_size_3(self):

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
            ]
        )

        mask = np.array([[False, True], [True, False]])

        sub_array_1d = al.array_mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=3
        )

        assert (
            sub_array_1d
            == np.array([1, 2, 3, 7, 8, 9, 13, 14, 15, 4, 5, 6, 10, 11, 12, 16, 17, 18])
        ).all()


class TestSubArray2dFromSubArray1d(object):
    def test__simple_2d_array__is_masked_and_mapped__sub_grid_size_1(self):

        array_1d = np.array([1.0, 2.0, 3.0, 4.0])

        mask = np.full(fill_value=False, shape=(2, 2))

        array_2d = al.array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=1
        )

        assert (array_2d == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

        array_1d = np.array([1.0, 2.0, 3.0])

        mask = np.array([[False, False], [False, True]])

        array_2d = al.array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=1
        )

        assert (array_2d == np.array([[1.0, 2.0], [3.0, 0.0]])).all()

        array_1d = np.array([1.0, 2.0, 3.0, -1.0, -2.0, -3.0])

        mask = np.array(
            [
                [False, False, True, True],
                [False, True, True, True],
                [False, False, True, False],
            ]
        )

        array_2d = al.array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=1
        )

        assert (
            array_2d
            == np.array(
                [[1.0, 2.0, 0.0, 0.0], [3.0, 0.0, 0.0, 0.0], [-1.0, -2.0, 0.0, -3.0]]
            )
        ).all()

    def test__simple_2d_array__is_masked_and_mapped__sub_grid_size_2(self):

        array_1d = np.array(
            [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 4.0]
        )

        mask = np.array([[False, False], [False, True]])

        array_2d = al.array_mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=2
        )

        assert (
            array_2d
            == np.array(
                [
                    [1.0, 1.0, 2.0, 2.0],
                    [1.0, 1.0, 2.0, 2.0],
                    [3.0, 3.0, 0.0, 0.0],
                    [3.0, 4.0, 0.0, 0.0],
                ]
            )
        ).all()



import autofit as af
import autolens as al

import os

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = workspace_path + "../../outputs/PyAutoLens/subhalo_challenge"

af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=output_path
)

data_type = "analytic_psf"

data_level = "level_0"

data_name = "large_hi_sn_system_1"
data_name = "small_hi_sn_system_1"

data_level = "level_1"
# data_name = 'large_hi_sn_system_1'
# data_name = 'large_hi_sn_system_2'
# data_name = 'large_md_sn_system_1'
# data_name = 'large_md_sn_system_2'
# data_name = 'large_lo_sn_system_1'
# data_name = 'large_lo_sn_system_2' ##
# data_name = 'small_hi_sn_system_1'
# data_name = 'small_hi_sn_system_2'
# data_name = 'small_md_sn_system_1'
# data_name = 'small_md_sn_system_2'
data_name = "small_lo_sn_system_1"
# data_name = 'small_lo_sn_system_2'

pixel_scale = 0.00976562

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path,
    folder_names=["data", "subhalo_challenge", data_type, data_level, data_name],
)

resized_shape = (700, 700)

ccd_data = al.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
    resized_ccd_shape=resized_shape,
    resized_psf_shape=(9, 9),
)

mask = al.load_mask_from_fits(
    mask_path=data_path + "mask_irregular.fits", pixel_scale=pixel_scale
)

mask = mask.resized_scaled_array_from_array(new_shape=resized_shape)

positions = al.load_positions(positions_path=data_path + "positions.dat")

pipeline_settings = al.PipelineSettingsHyper(
    hyper_galaxies=True,
    hyper_image_sky=False,
    hyper_background_noise=True,
    include_shear=True,
    pixelization=al.pixelizations.VoronoiBrightnessImage,
    regularization=al.regularization.AdaptiveBrightness,
)

from workspace_jam.pipelines.hyper.no_lens_light.initialize import (
    lens_sie__source_sersic,
)
from workspace_jam.pipelines.hyper.no_lens_light.inversion.from_initialize import (
    lens_sie__source_inversion,
)
from workspace_jam.pipelines.hyper.no_lens_light.power_law.from_inversion import (
    lens_power_law__source_inversion,
)

from workspace_jam.pipelines.hyper.no_lens_light.subhalo.from_power_law import (
    lens_power_law__subhalo_nfw__source_inversion,
)

pipeline_initialize = lens_sie__source_sersic.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_type, data_level, data_name],
    positions_threshold=1.0,
)

pipeline_inversion = lens_sie__source_inversion.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_type, data_level, data_name],
    positions_threshold=1.0,
    inversion_pixel_limit=800,
)

pipeline_power_law = lens_power_law__source_inversion.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_type, data_level, data_name],
    positions_threshold=0.3,
    inversion_pixel_limit=800,
)

pipeline_subhalo = lens_power_law__subhalo_nfw__source_inversion.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_type, data_level, data_name],
    positions_threshold=0.3,
    inversion_pixel_limit=800,
    parallel=True,
)

pipeline = (
    pipeline_initialize + pipeline_inversion + pipeline_power_law + pipeline_subhalo
)

pipeline.run(data=ccd_data, mask=mask, positions=positions)



import autolens as al
from test.integration.tests.promise import lens_mass__source_x2
from test.integration.tests.runner import run_a_mock


class TestCase:
    def test_lens_mass__source_x2(self):
        run_a_mock(lens_mass__source_x2)



from autolens.data.instrument import abstract_data
from autolens.array import grids
from autolens.lens import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

import matplotlib.pyplot as plt
from workspace_jam.helping.qiuhan.deflection_map import deflection_multi_map
from astropy.cosmology import FlatLambdaCDM


pixel_scale_img = 0.05
npixel_img = int(10 / pixel_scale_img)

H0 = 100.0
OM0 = 0.3
cosmo = FlatLambdaCDM(H0=H0, Om0=OM0)

psf = abstract_data.PSF.from_gaussian(
    shape=(11, 11), sigma=0.1, pixel_scale=pixel_scale_img
)

image_plane_grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=(200, 200), pixel_scale=pixel_scale_img, psf_shape=psf.shape, sub_grid_size=1
)

lens_galaxy = g.Galaxy(
    mass=al.mass_profiles.SphericalIsothermal(centre=(3.0, 3.0), einstein_radius=1.6),
    redshift=0.5,
)

source_galaxy = g.Galaxy(
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.0),
        axis_ratio=0.8,
        phi=0.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
    redshift=1.0,
)

# "zero kappa" Mass Sheet
mass_sheet = g.Galaxy(
    sheet=al.mass_profiles.MassSheet(centre=(0.0, 0.0), kappa=0.0), redshift=0.2
)


# Below, all variables with "_test" means inpluding the mass sheet and all
# variables with "_ref" means they don't have a mass sheet

lens_list_ref = [lens_galaxy, source_galaxy]

tracer_ref = ray_tracing.Tracer(
    galaxies=lens_list_ref, image_plane_grid=image_plane_grid, cosmology=cosmo
)

################################################################################
# Here, I use deflection_multi_map function to calculate the deflection angles
# in multiplane situations. This function will return an N*N*2 array with
# [:, :, 1] to be deflections in x direction and [:, :, 0] to be deflections in
# y direction.

alpha_ref = deflection_multi_map(
    N=npixel_img, map_scale=pixel_scale_img, tracer=tracer_ref, z_lens=0.5, z_source=1.0
)

alphax_ref = alpha_ref[:, :, 1]
alphay_ref = alpha_ref[:, :, 0]

lens_list_test = [lens_galaxy, mass_sheet, source_galaxy]

tracer_test = ray_tracing.Tracer(
    galaxies=lens_list_test, image_plane_grid=image_plane_grid, cosmology=cosmo
)

alpha_test = deflection_multi_map(
    N=npixel_img,
    map_scale=pixel_scale_img,
    tracer=tracer_test,
    z_lens=0.5,
    z_source=1.0,
)

alphax_test = alpha_test[:, :, 1]
alphay_test = alpha_test[:, :, 0]


fig, axes = plt.subplots(1, 2, figsize=(8, 4))
axes[0].imshow(alphax_ref)
axes[0].set_title("no mass sheet")
axes[1].imshow(alphax_test)
axes[1].set_title("with mass sheet")
plt.show()



# Up to now, we've not paid much attention to the source galaxy's morphology. We've assumed its a single-component
# exponential profile, which is a fairly crude assumption. A quick look at any image of a real galaxy reveals a wealth
# of different structures that could be present - bulges, disks, bars, star-forming knots and so on.
# Furthermore, there could be more than one source-galaxy!

# In this example, we'll explore how far we can get trying to_fit a complex source using a pipeline. Fitting complex
# source's is an exercise in diminishing returns. Each component we add to our source model brings with it an extra 5-7,
# parameters. If there are 4 components, or multiple galaxies, we're quickly entering the somewhat nasty regime of
# 30-40+ parameters in our non-linear search. Even with a pipeline, that is a lot of parameters to fit!

### AUTOFIT + CONFIG SETUP ###

import autofit as af

# Setup the path to the workspace, using by filling in your path below.
workspace_path = "/path/to/user/autolens_workspace/"
workspace_path = "/home/jammy/PycharmProjects/PyAutoLens/workspace/"

# Setup the path to the config folder, using the workspace path.
config_path = workspace_path + "config"

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=config_path, output_path=workspace_path + "output"
)

### AUTOLENS + DATA SETUP ###

import autolens as al

# This function simulates an image with a complex source.
def simulate():

    psf = al.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    grid = al.Grid.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05
    )

    lens_galaxy = al.Galaxy(
        redshift=0.5,
        mass=al.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy_0 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=1.0,
            sersic_index=1.5,
        ),
    )

    source_galaxy_1 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(-0.25, 0.25),
            axis_ratio=0.7,
            phi=45.0,
            intensity=0.1,
            effective_radius=0.2,
            sersic_index=3.0,
        ),
    )

    source_galaxy_2 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(0.45, -0.35),
            axis_ratio=0.6,
            phi=90.0,
            intensity=0.03,
            effective_radius=0.3,
            sersic_index=3.5,
        ),
    )

    source_galaxy_3 = al.Galaxy(
        redshift=1.0,
        light=al.light_profiles.EllipticalSersic(
            centre=(-0.05, -0.0),
            axis_ratio=0.9,
            phi=140.0,
            intensity=0.03,
            effective_radius=0.1,
            sersic_index=4.0,
        ),
    )

    tracer = al.Tracer.from_galaxies(
        galaxies=[
            lens_galaxy,
            source_galaxy_0,
            source_galaxy_1,
            source_galaxy_2,
            source_galaxy_3,
        ]
    )

    return al.SimulatedCCDData.from_tracer_grid_and_exposure_arrays(
        tracer=tracer,
        grid=grid,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Lets Simulate the CCD data.
ccd_data = simulate()
al.ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Yep, that's a pretty complex source. There are clearly more than 4 peaks of light - I wouldn't like to guess whether
# there is one or two sources (or more). You'll also notice I omitted the lens galaxy's light for this system. This is
# to keep the number of parameters down and the phases running fast, but we wouldn't get such a luxury for a real galaxy.

# Again, before we checkout the pipeline, lets import it, and get it running.
from workspace.howtolens.chapter_3_pipelines import tutorial_3_pipeline_complex_source

pipeline_complex_source = tutorial_3_pipeline_complex_source.make_pipeline(
    phase_folders=["howtolens", "c3_t3_complex_source"]
)

pipeline_complex_source.run(data=ccd_data)

# Okay, so with 4 sources, we still couldn't get a good a fit to the source that didn't leave residuals. However, I actually
# simulated the lens with 4 sources. This means that there is a 'perfect fit' somewhere in parameter
# space that we unfortunately missed using the pipeline above.

# Lets confirm this, by manually fitting the ccd data with the true input model.

lens_data = al.LensData(
    ccd_data=ccd_data,
    mask=al.Mask.circular(
        shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0
    ),
)

lens_galaxy = al.Galaxy(
    redshift=0.5,
    mass=al.mass_profiles.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
    ),
)

source_galaxy_0 = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=90.0,
        intensity=0.2,
        effective_radius=1.0,
        sersic_index=1.5,
    ),
)

source_galaxy_1 = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(-0.25, 0.25),
        axis_ratio=0.7,
        phi=45.0,
        intensity=0.1,
        effective_radius=0.2,
        sersic_index=3.0,
    ),
)

source_galaxy_2 = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(0.45, -0.35),
        axis_ratio=0.6,
        phi=90.0,
        intensity=0.03,
        effective_radius=0.3,
        sersic_index=3.5,
    ),
)

source_galaxy_3 = al.Galaxy(
    redshift=1.0,
    light=al.light_profiles.EllipticalSersic(
        centre=(-0.05, -0.0),
        axis_ratio=0.9,
        phi=140.0,
        intensity=0.03,
        effective_radius=0.1,
        sersic_index=4.0,
    ),
)

tracer = al.Tracer.from_galaxies(
    galaxies=[
        lens_galaxy,
        source_galaxy_0,
        source_galaxy_1,
        source_galaxy_2,
        source_galaxy_3,
    ]
)

true_fit = al.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

al.lens_fit_plotters.plot_fit_subplot(
    fit=true_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

al.lens_fit_plotters.plot_fit_subplot_of_planes(
    fit=true_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

# And indeed, we see an improved residual-map, chi-squared-mao, and so forth.

# The morale of this story is that if the source morphology is complex, there is no way we can build a pipeline to
# fit it. For this tutorial, this was true even though our source model could actually fit the data perfectly. For
# real lenses, the source will be *even more complex* and there is even less hope of getting a good fit :(

# But fear not, PyAutoLens has you covered. In chapter 4, we'll introduce a completely new way to model the source
# galaxy, which addresses the problem faced here. But before that, in the next tutorial we'll discuss how we actually
# pass priors in a pipeline.
